/* $Id $
	------------------------------------------------------------------
	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	@author Nicklas
*/

var MoveBioMaterial = function()
{
	var move = {};
	
	var graphics;
	var pen;
	var selectedPen;
	
	// Source and destination plates
	var sourcePlate;
	var destPlate;
	
	// We can select one well on each plate
	var selectedSourceWell;
	var selectedDestWell;

	/**
		Initialize the page.
	*/
	move.initPage = function()
	{
		// Save and close buttons
		Buttons.addClickHandler('close', App.closeWindow);
		Buttons.addClickHandler('btnSave', move.moveBioMaterial);
		
		// Protocol
		Buttons.addClickHandler('protocol_id.select', Wizards.selectProtocol);
		Events.addEventHandler('protocol_id', 'base-selected', Wizards.setProtocolCallback);

		// Kit
		Buttons.addClickHandler('kit_id.select', Wizards.selectKit);
		Events.addEventHandler('kit_id', 'base-selected', Wizards.setKitCallback);

		// Hardware
		Buttons.addClickHandler('hardware_id.select', Wizards.selectHardware);
		Events.addEventHandler('hardware_id', 'base-selected', Wizards.setHardwareCallback);
		
		// Toolbar
		Buttons.addClickHandler('btnSelectPlate', move.selectBioPlate);
		Events.addEventHandler('btnSelectPlate', 'base-selected', move.setBioPlateCallback);
		Buttons.addClickHandler('btnPlaceByRow', move.placeByRow);
		Buttons.addClickHandler('btnPlaceByColumn', move.placeByColumn);
		Buttons.addClickHandler('btnClearMapping', move.clearMapping);
		Buttons.addClickHandler('btnSelectPlateMapping', move.selectPlateMapping);
		Events.addEventHandler('btnSelectPlateMapping', 'base-selected', move.setPlateMappingCallback);
		
		// Options
		Events.addEventHandler('showSourceCoordinates', 'click', move.showSourceCoordinatesOnClick);
		
		// Handle layout changes
		Events.addEventHandler(window, 'resize', move.onLayoutChange);

		// Initialize graphics
		graphics = new jsGraphics(Doc.element('canvas'));
		pen = new jsPen(new jsColor('#2288AA'), 1);
		selectedPen = new jsPen(new jsColor('#2288AA'), 2);
		
		move.displayBioPlate(Data.int('page-data', 'source-plate'), true);
	}
	
	move.selectBioPlate = function()
	{
		if (destPlate && !confirm('This will reset all moved biomaterial. Continue?')) return;
		
		var url = '&resetTemporary=1';
		var bioMaterialType = Data.int('page-data', 'biomaterial-type');
		if (bioMaterialType)
		{
			url += '&tmpfilter:INT:bioPlateType.bioMaterialType='+encodeURIComponent('|'+bioMaterialType);
		}
		var commonSubtype = Data.int('page-data', 'common-subtype');
		if (commonSubtype)
		{
			url += '&tmpfilter:INT:bioPlateType.itemSubtype='+encodeURIComponent('|'+commonSubtype);
		}

		Dialogs.selectItem('BIOPLATE', 'btnSelectPlate', 0, url);
	}

	move.setBioPlateCallback =  function(event)
	{
		move.displayBioPlate(event.detail.id, false);
	}
	
	move.displayBioPlate = function(bioPlateId, isSourcePlate)
	{
		if (!isSourcePlate && bioPlateId == sourcePlate.id)
		{		
			Forms.showNotification('plate.dest.name', 'It is not possible to use the same plate as both source and destination plate.');
			return false;
		}
		
		var request = Ajax.getXmlHttpRequest();
		var url = '../bioplates/ajax.jsp?ID='+App.getSessionId()+'&cmd=GetFullPlateInfo&item_id=' + bioPlateId;
		request.open("GET", url, false); 
		// NOTE! 'false' causes code to wait for the response. aka. 'Synchronous AJAX' or SJAX.
		request.send(null);
		var plateInfo = JSON.parse(request.responseText);
		if (plateInfo.status != 'ok')
		{
			Forms.showNotification('plate.dest.name', plateInfo.message);
			return;
		}

		// Get plate and well information from the AJAX response
		var rows = plateInfo.rows
		var columns = plateInfo.columns;
		
		var bigPlate = rows > 12 || columns > 18;
		var plateClass = bigPlate ? 'plate bigplate' : 'plate';

		// Remove existing mapping
		if (sourcePlate) sourcePlate.unmapAll(graphics);
		selectedDestWell = null;

		// Create plate and wells
		var prefix = isSourcePlate ? 'source' : 'dest';
		var plate = new Plate(prefix, plateInfo.id, plateInfo.name, rows, columns, isSourcePlate);
		for (var i = 0; i < plateInfo.wells.length; i++)
		{
			var wellInfo = plateInfo.wells[i];
			var bmInfo = wellInfo.bioMaterial;
			var row = wellInfo.row;
			var col = wellInfo.column;
			var well = plate.getWell(row, col);
			if (bmInfo)
			{
				well.id = bmInfo.id;
				well.name = bmInfo.name;
			}
			well.locked = isSourcePlate ? !wellInfo.canClear : !wellInfo.canAdd || well.id;
		}
		
		// Create html table representing the bioplate
		var html = '<table class="'+plateClass+'">';
		html += '<tr><td class="bg-filled-100"></td>';
		for (var c = 0; c < plate.columns; c++)
		{
			html += '<td class="columnheader bg-filled-100">' + (c+1) + '</td>';
		}
		html += '</tr>';
		for (var r = 0; r < plate.rows; r++)
		{
			html += '<tr><td class="rowheader bg-filled-100">' + Plates.toAlphaCoordinate[r] + '</td>';
			for (var c = 0; c < plate.columns; c++)
			{
				var well = plate.getWell(r, c);
				var cls = 'well';
				var title = well.name ? Strings.encodeTags(well.name) : '';
				if (isSourcePlate)
				{
					// We can only move a biomaterial if the well is not empty or locked
					if (well.id)
					{
						cls += well.locked ? ' used locked' : ' used editable';
					}
					else
					{
						cls += ' empty';
					}
				}
				else
				{
					// We only add biomaterial to destination well if it is not locked
					if (well.locked)
					{
						cls += ' unmappable';
					}
					else
					{
						cls += ' empty editable';
					}
				}
				html += '<td id="'+prefix+'.'+r+'.'+c+'" class="' + cls + '" data-row="'+r+'" data-column="'+c+'" title="'+title+'">';
				html += '<div class="info" id="'+prefix+'.'+r+'.'+c+'.info"></div></td>';
			}
			html += '</tr>';
		}
		html += '</table>';
		
		if (isSourcePlate)
		{
			Doc.element('plate.src').innerHTML = html;
			Doc.element('plate.src.name').innerHTML = Strings.encodeTags(plateInfo.name);
			sourcePlate = plate;
			
			var wells = Doc.element('plate.src').getElementsByClassName('editable');
			for (var i = 0; i < wells.length; i++)
			{
				var well = wells[i];
				Events.addEventHandler(well, 'click', move.sourceWellOnClick);
				Events.addEventHandler(well, 'mouseout', move.sourceWellOnMouseOut);
				Events.addEventHandler(well, 'mouseover', move.sourceWellOnMouseOver);
			}
		}
		else
		{
			Doc.element('plate.dest').innerHTML = html;
			Doc.element('plate.dest.name').innerHTML = Strings.encodeTags(plateInfo.name);
			destPlate = plate;
			Doc.showHide('plate.dest.options', !bigPlate);
			
			var wells = Doc.element('plate.dest').getElementsByClassName('editable');
			for (var i = 0; i < wells.length; i++)
			{
				var well = wells[i];
				Events.addEventHandler(well, 'click', move.destWellOnClick);
				Events.addEventHandler(well, 'mouseout', move.destWellOnMouseOut);
				Events.addEventHandler(well, 'mouseover', move.destWellOnMouseOver);
			}
			
			move.showSourceCoordinatesOnClick();
		}
	}

	move.moveBioMaterial = function()
	{
		if (!destPlate)
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		var frm = document.forms['main'];
		if (Strings.trim(frm.name.value) == '')
		{
			Forms.showNotification(frm.name, 'You must enter a name for the event');
			return;
		}
		
		frm.destplate_id.value = destPlate.id;
		frm.rows.value = destPlate.rows;
		frm.columns.value = destPlate.columns;
		
		var numMapped = 0;
		for (var row = 0; row < destPlate.rows; row++)
		{
			for (var column = 0; column < destPlate.columns; column++)
			{
				var well = destPlate.getWell(row, column);
				
				if (well.mappedWell)
				{
					Forms.addHidden(frm, 'well.' + row + '.' + column, well.mappedWell.id);
					numMapped++;
				}
			}
		}
		
		if (numMapped == 0)
		{
			Forms.showNotification('btnSave', 'No wells have been mapped.');
			return;
		}

		frm.submit();
	}
	
	/**
		Select the source well that is clicked on. The currently selected well
		is de-selected. If the clicked well is the same as the currently 
		selected well no new item is selected. If a destination well is already 
		selected a link is made between the source and destination wells.
	*/
	move.sourceWellOnClick = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');
		App.debug(row+':'+column);
		var well = sourcePlate.getWell(row, column);
		if (!well) return;
	
		// De-select the currently selected source well
		if (selectedSourceWell) 
		{
			selectedSourceWell.setSelected(false);
			if (well == selectedSourceWell)
			{
				// Re-draw link with regular pen (eg. same as onmouseover)
				selectedSourceWell.drawLink(graphics, pen, true);
				selectedSourceWell = null;
				return;
			}
			else
			{
				selectedSourceWell.hideLink(graphics);
			}
		}
		
		// Select the new source well and draw link to destination well
		selectedSourceWell = well;
		selectedSourceWell.setSelected(true);
		selectedSourceWell.drawLink(graphics, selectedPen, true);
	
		// Map the source and destination wells
		if (selectedSourceWell && selectedDestWell)
		{
			move.mapSelectedWells();
		}
	}
	
	/**
		Draw a link between the mapped source and destination wells.
	*/
	move.sourceWellOnMouseOver = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');

		var well = sourcePlate.getWell(row, column);
		if (!well) return;
		well.drawLink(graphics, pen, false);
	}
	
	/**
		Hide the link between the mapped source and destination 
		wells, unless one of them is selected.
	*/
	move.sourceWellOnMouseOut = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');

		var well = sourcePlate.getWell(row, column);
		if (!well) return;
		if (!well.selected && !(well.mappedWell && well.mappedWell.selected)) 
		{
			well.hideLink(graphics);
		}
	}

	move.destWellOnClick = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');

		var well = destPlate.getWell(row, column);
		// De-select the currently selected well
		if (selectedDestWell) 
		{
			selectedDestWell.setSelected(false);
			if (well == selectedDestWell)
			{
				// Re-draw link with regular pen (eg. same as onmouseover)
				selectedDestWell.drawLink(graphics, pen, true);
				selectedDestWell = null;
				return;
			}
			else
			{
				selectedDestWell.hideLink(graphics);
			}
		}
		
		selectedDestWell = well;
		selectedDestWell.setSelected(true);
		selectedDestWell.drawLink(graphics, selectedPen, true);

		if (selectedSourceWell && selectedDestWell)
		{
			move.mapSelectedWells();
		}
	}
	
	move.destWellOnMouseOver = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');

		var well = destPlate.getWell(row, column);
		if (!well) return;
		well.drawLink(graphics, pen, false);
	}
	
	move.destWellOnMouseOut = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'column');

		var well = destPlate.getWell(row, column);
		if (!well) return;
		if (!well.selected && !(well.mappedWell && well.mappedWell.selected)) 
		{
			well.hideLink(graphics);
		}
	}
	
	/**
		Map the selected source and destination wells. Hide and
		redraw links as needed.
	*/
	move.mapSelectedWells = function()
	{
		// Hide any links that are currently displayed
		selectedSourceWell.hideLink(graphics);
		selectedDestWell.hideLink(graphics);
		
		// Map to the new well and draw link
		if (selectedSourceWell.mappedWell != selectedDestWell)
		{
			selectedSourceWell.mapToWell(selectedDestWell);
			selectedSourceWell.drawLink(graphics, pen, true);
		}
		else
		{
			selectedSourceWell.unmapWell();
		}
		
		// De-select everything
		selectedSourceWell.setSelected(false);
		selectedDestWell.setSelected(false);
		selectedSourceWell = null;
		selectedDestWell = null;
	}

	/**
		Remove all mappings that have been made so far.
	*/
	move.clearMapping = function()
	{
		if (!destPlate) 
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		if (!confirm('This will remove all placed biomaterial. Continue?')) return;
		sourcePlate.unmapAll(graphics);
		if (selectedSourceWell)
		{
			selectedSourceWell.setSelected(false);
			selectedSourceWell = null;
		}
		if (selectedDestWell)
		{
			selectedDestWell.setSelected(false);
			selectedDestWell = null;
		}
	}

	/**
		Automatically move remaining biomaterial to the destination plate filling rows first.
		Biomaterial and wells that have already been mapped are skipped.
	*/
	move.placeByRow = function()
	{
		if (!destPlate) 
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		var srcRow = 0;
		var srcCol = 0;
		for (var destRow = 0; destRow < destPlate.rows; destRow++)
		{
			for (var destCol = 0; destCol < destPlate.columns; destCol++)
			{
				var destWell = destPlate.getWell(destRow, destCol);
				if (!destWell.mappedWell && !destWell.locked)
				{
					var mapped = false;
					var srcWell = sourcePlate.getWell(srcRow, srcCol);
					while (srcWell && !mapped)
					{
						if (srcWell.id && !srcWell.locked && !srcWell.mappedWell)
						{
							srcWell.mapToWell(destWell);
							mapped = true;
						}
						else
						{
							srcCol++;
							if (srcCol >= sourcePlate.columns)
							{
								srcCol = 0;
								srcRow++;
							}
							srcWell = sourcePlate.getWell(srcRow, srcCol);
						}
					}
				}
			}
		}
	}
	
	/**
		Automatically move remaining biomaterial to the destination plate filling columns first.
		Biomaterial and wells that have already been mapped are skipped.
	*/
	move.placeByColumn = function()
	{
		if (!destPlate) 
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		var srcRow = 0;
		var srcCol = 0;
		for (var destCol = 0; destCol < destPlate.columns; destCol++)
		{
			for (var destRow = 0; destRow < destPlate.rows; destRow++)
			{
				var destWell = destPlate.getWell(destRow, destCol);
				if (!destWell.mappedWell && !destWell.locked)
				{
					var mapped = false;
					var srcWell = sourcePlate.getWell(srcRow, srcCol);
					while (srcWell && !mapped)
					{
						if (srcWell.id && !srcWell.locked && !srcWell.mappedWell)
						{
							srcWell.mapToWell(destWell);
							mapped = true;
						}
						else
						{
							srcRow++;
							if (srcRow >= sourcePlate.rows)
							{
								srcCol++;
								srcRow = 0;
							}
							srcWell = sourcePlate.getWell(srcRow, srcCol);
						}
					}
				}
			}
		}
	}

	/**
		Redraw the link between selected wells when the layout changes
	*/
	move.onLayoutChange = function(event)
	{
		if (selectedSourceWell)
		{
			selectedSourceWell.drawLink(graphics, selectedPen, true);
		}
		else if (selectedDestWell)
		{
			selectedDestWell.drawLink(graphics, selectedPen, true);
		}
	}

	move.showSourceCoordinatesOnClick = function()
	{
		var frm = document.forms['main'];
		Doc.addOrRemoveClass('plate.dest', 'noinfo', !frm.showSourceCoordinates.checked);
	}

	move.selectPlateMapping = function()
	{
		if (!destPlate)
		{
			Forms.showNotification('btnSelectPlate', 'No destination plate has been selected');
			return;
		}
		var url = '&resetTemporary=1';
		url += '&tmpfilter:INT:sourceGeometry.rows='+sourcePlate.rows;
		url += '&tmpfilter:INT:sourceGeometry.columns='+sourcePlate.columns;
		url += '&tmpfilter:INT:sourceCount=1';
		url += '&tmpfilter:INT:destinationGeometry.rows='+destPlate.rows;
		url += '&tmpfilter:INT:destinationGeometry.columns='+destPlate.columns;
		url += '&tmpfilter:INT:destinationCount=1';
		Dialogs.selectItem('PLATEMAPPING', 'btnSelectPlateMapping', 0, url);
	}

	move.setPlateMappingCallback = function(event)
	{
		var request = Ajax.getXmlHttpRequest();
		var url = '../../lims/platemappings/ajax.jsp?ID='+App.getSessionId();
		url += '&cmd=GetMappingDetails&item_id=' + event.detail.id;
		request.open("GET", url, false);
		request.send(null);
		var mappingInfo = JSON.parse(request.responseText);
		if (mappingInfo.status != 'ok')
		{
			Forms.showNotification('btnSelectPlateMapping', plateInfo.message);
			return false;
		}
		
		var win = Dialogs.getDialog('SelectPLATEMAPPING') || window;
		if (mappingInfo.sourcePlates != 1 || mappingInfo.sourceRows != sourcePlate.rows || mappingInfo.sourceColumns != sourcePlate.columns)
		{
			if (!win.confirm("The selected mapping doesn't match the geometry of the source plate.\nContinue mapping overlapping positions?")) return;
		}
		else if (mappingInfo.destinationPlates != 1 || mappingInfo.destinationRows != destPlate.rows || mappingInfo.destinationColumns != destPlate.columns)
		{
			if (!win.confirm("The selected mapping doesn't match the geometry of the destination plate.\nContinue mapping overlapping positions?")) return;
		}
		
		var numMapped = 0;
		for (var i = 0; i < mappingInfo.details.length; i++)
		{
			var mapping = mappingInfo.details[i].split(',');
			var srcWell = sourcePlate.getWell(parseInt(mapping[1]), parseInt(mapping[2]));
			var destWell = destPlate.getWell(parseInt(mapping[4]), parseInt(mapping[5]));
			if (srcWell && destWell)
			{
				if (!destWell.mappedWell && !destWell.locked && srcWell.id && !srcWell.locked && !srcWell.mappedWell)
				{
					srcWell.mapToWell(destWell);
					numMapped++;
				}
			}
		}
	}
	
	return move;
}();

Doc.onLoad(MoveBioMaterial.initPage);
